/**
 * Copyright (c) 2017, Łukasz Marcin Podkalicki <lpodkalicki@gmail.com>
 * ESP32/016
 * WiFi Sniffer.
 */

#include "freertos/FreeRTOS.h"
#include "esp_wifi.h"
#include "esp_wifi_types.h"
#include "esp_system.h"
#include "esp_event.h"
#include "esp_event_loop.h"
#include "nvs_flash.h"
#include "driver/gpio.h"
#include <string.h>

#define	LED_GPIO_PIN			GPIO_NUM_4
#define	WIFI_CHANNEL_MAX		(11)
#define	WIFI_CHANNEL_SWITCH_INTERVAL	(500)
#define	RSSI_AVERAGES		(32)

static wifi_country_t wifi_country = {.cc="CA", .schan=1, .nchan=11, .policy=WIFI_COUNTRY_POLICY_AUTO};

typedef struct {
	unsigned frame_ctrl:16;
	unsigned duration_id:16;
	uint8_t addr1[6]; /* receiver address */
	uint8_t addr2[6]; /* sender address */
	uint8_t addr3[6]; /* filtering address */
	unsigned sequence_ctrl:16;
	uint8_t addr4[6]; /* optional */
} wifi_ieee80211_mac_hdr_t;

typedef struct {
	wifi_ieee80211_mac_hdr_t hdr;
	uint8_t payload[0]; /* network data ended with 4 bytes csum (CRC32) */
} wifi_ieee80211_packet_t;

typedef struct {
	uint8_t addr[6]; /* sender address */
	int32_t rssi[RSSI_AVERAGES]; /* RSSI */
	uint8_t set; /* Set if fields are populated */
	int32_t points; /* Sample point quantity */
	int32_t avg; /* Averaged */
} wifi_ieee80211_pwr_t;

static wifi_ieee80211_pwr_t rssi_rolling[20];

static esp_err_t event_handler(void *ctx, system_event_t *event);
static void wifi_sniffer_init(void);
static void wifi_sniffer_report(void);
static void wifi_sniffer_set_channel(uint8_t channel);
static const char *wifi_sniffer_packet_type2str(wifi_promiscuous_pkt_type_t type);
static void wifi_sniffer_packet_handler(void *buff, wifi_promiscuous_pkt_type_t type);

void
app_main(void)
{
	uint8_t level = 0, channel = 1, interval = 0;

	/* setup */
	wifi_sniffer_init();
	gpio_set_direction(LED_GPIO_PIN, GPIO_MODE_OUTPUT);

	/* loop */
	while (true) {
		gpio_set_level(LED_GPIO_PIN, level ^= 1);
		vTaskDelay(WIFI_CHANNEL_SWITCH_INTERVAL / portTICK_PERIOD_MS);
		wifi_sniffer_set_channel(channel);
		channel = (channel % WIFI_CHANNEL_MAX) + 1;
        
        if((interval % 1) == 0) {
            printf("Report:\n");
			wifi_sniffer_report();
        }
        
        interval++;
    }
}

esp_err_t
event_handler(void *ctx, system_event_t *event)
{
	
	return ESP_OK;
}

void
wifi_sniffer_init(void)
{

	nvs_flash_init();
    	tcpip_adapter_init();
    	ESP_ERROR_CHECK( esp_event_loop_init(event_handler, NULL) );
    	wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
	ESP_ERROR_CHECK( esp_wifi_init(&cfg) );
	ESP_ERROR_CHECK( esp_wifi_set_country(&wifi_country) ); /* set country for channel range [1, 13] */
	ESP_ERROR_CHECK( esp_wifi_set_storage(WIFI_STORAGE_RAM) );
    	ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_NULL) );
    	ESP_ERROR_CHECK( esp_wifi_start() );
	esp_wifi_set_promiscuous(true);
	esp_wifi_set_promiscuous_rx_cb(&wifi_sniffer_packet_handler);
}

void
wifi_sniffer_set_channel(uint8_t channel)
{
	
	esp_wifi_set_channel(channel, WIFI_SECOND_CHAN_NONE);
}

void
wifi_sniffer_report()
{
	uint8_t i;
	for (i = 0; i < 20; i++) {
		if(rssi_rolling[i].set == 0) {
			//printf("Report finished\n");
			return;
		}
			
		printf("ADDR=%02x:%02x:%02x:%02x:%02x:%02x, PTS=%05d PWR=%02d\n", 
			rssi_rolling[i].addr[0],rssi_rolling[i].addr[1],rssi_rolling[i].addr[2],
			rssi_rolling[i].addr[3],rssi_rolling[i].addr[4],rssi_rolling[i].addr[5],
			rssi_rolling[i].points, rssi_rolling[i].avg);
	}
} 

const char *
wifi_sniffer_packet_type2str(wifi_promiscuous_pkt_type_t type)
{
	switch(type) {
	case WIFI_PKT_MGMT: return "MGMT";
	case WIFI_PKT_DATA: return "DATA";
	default:	
	case WIFI_PKT_MISC: return "MISC";
	}
}

unsigned char wifi_sniffer_filter(wifi_promiscuous_pkt_type_t *type, const wifi_ieee80211_mac_hdr_t *hdr)
{
    //printf("PACKET TYPE=%s\n", wifi_sniffer_packet_type2str(*type));
    /*printf("ADDR1=%02x:%02x:%02x:%02x:%02x:%02x\n", 
        hdr->addr1[0],hdr->addr1[1],hdr->addr1[2],
		hdr->addr1[3],hdr->addr1[4],hdr->addr1[5]);
        */
    if (*type == WIFI_PKT_MGMT)
		return 0;
    
    unsigned char equal = 1;
    for (unsigned char i = 0; i < 6; i++) {
        if (hdr->addr1[i] != 0xff) {
            equal = 0;
        }
    }
    if (equal == 1) {
        return 0;
    }
    
    return 1;
}

void wifi_sniffer_collate(const wifi_ieee80211_mac_hdr_t *hdr, int32_t rssi_current)
{

	uint8_t i, a, equal;
	int rssi_total = 0;
	for (i = 0; i < 20; i++) {
		equal = 1;
		for (a = 0; a < 6; a++) {
			if (hdr->addr1[a] != rssi_rolling[i].addr[a]) {
				equal = 0;
			}
		}
		if ((equal == 1) || (rssi_rolling[i].set == 0)) {
			if (rssi_rolling[i].set == 0) {
				rssi_rolling[i].points = 0;
				rssi_rolling[i].set = 1;
			}
			for (a = 0; a < 6; a++) {
				rssi_rolling[i].addr[a] = hdr->addr1[a];
			}
			for (a = 0; a < RSSI_AVERAGES - 1; a++) {
				rssi_rolling[i].rssi[a] = rssi_rolling[i].rssi[a + 1];
			}
			rssi_rolling[i].rssi[RSSI_AVERAGES - 1] = rssi_current;

			rssi_rolling[i].points++;
            int32_t samples = RSSI_AVERAGES;
			if (rssi_rolling[i].points < RSSI_AVERAGES) {
                samples = rssi_rolling[i].points;
            }
            
			for (a = 1; a <= samples; a++) {
				rssi_total += rssi_rolling[i].rssi[RSSI_AVERAGES - a];
			}
			rssi_rolling[i].avg = rssi_total / samples;
			return;

		} 
	}
    
    return;
}

void
wifi_sniffer_packet_handler(void* buff, wifi_promiscuous_pkt_type_t type)
{


	const wifi_promiscuous_pkt_t *ppkt = (wifi_promiscuous_pkt_t *)buff;
	const wifi_ieee80211_packet_t *ipkt = (wifi_ieee80211_packet_t *)ppkt->payload;
	const wifi_ieee80211_mac_hdr_t *hdr = &ipkt->hdr;
    
	if (!wifi_sniffer_filter(&type, hdr))
		return;
	wifi_sniffer_collate(hdr, ppkt->rx_ctrl.rssi);

	printf("PACKET TYPE=%s, CHAN=%02d, RSSI=%02d,"
		" ADDR1=%02x:%02x:%02x:%02x:%02x:%02x,"
		" ADDR2=%02x:%02x:%02x:%02x:%02x:%02x,"
		" ADDR3=%02x:%02x:%02x:%02x:%02x:%02x\n",
		wifi_sniffer_packet_type2str(type),
		ppkt->rx_ctrl.channel,
		ppkt->rx_ctrl.rssi,
		/* ADDR1 */
		hdr->addr1[0],hdr->addr1[1],hdr->addr1[2],
		hdr->addr1[3],hdr->addr1[4],hdr->addr1[5],
		/* ADDR2 */
		hdr->addr2[0],hdr->addr2[1],hdr->addr2[2],
		hdr->addr2[3],hdr->addr2[4],hdr->addr2[5],
		/* ADDR3 */
		hdr->addr3[0],hdr->addr3[1],hdr->addr3[2],
		hdr->addr3[3],hdr->addr3[4],hdr->addr3[5]
	);
}
